@@ -1,6 +1,14 @@
-from . import models
-from . import fields
+from __future__ import annotations
+
 import re
+from typing import TYPE_CHECKING, Any, Dict, Union
+
+from . import consts, fields, models
+
+if TYPE_CHECKING:
+    import xml.etree.ElementTree as ET
+
+    from .fattura_semplificata import FatturaElettronicaSemplificata
 
 #
 # This file describes the data model of the Italian Fattura Elettronica.
@@ -14,6 +22,17 @@ import re
 # class as a field, automatically wraps it in a `ModelField`.
 #
 
+__all__ = []
+
+
+def export(el):
+    """
+    Add a symbol to __all__ for export
+    """
+    global __all__
+    __all__.append(el.__name__)
+    return el
+
 
 NS = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.2"
 NS10 = "http://ivaservizi.agenziaentrate.gov.it/docs/xsd/fatture/v1.0"
@@ -70,24 +89,29 @@ class FullNameMixin:
                             " and ".join(x.name for x in should_not_be_set)))
 
 
+@export
 class IdFiscale(models.Model):
     id_paese = fields.StringField(length=2)
     id_codice = fields.StringField(max_length=28)
 
 
+@export
 class IdTrasmittente(IdFiscale):
     pass
 
 
+@export
 class IdFiscaleIVA(IdFiscale):
     pass
 
 
+@export
 class ContattiTrasmittente(models.Model):
     telefono = fields.StringField(min_length=5, max_length=12, null=True)
     email = fields.StringField(min_length=7, max_length=256, null=True)
 
 
+@export
 class DatiTrasmissione(models.Model):
     id_trasmittente = IdTrasmittente
     progressivo_invio = fields.ProgressivoInvioField()
@@ -134,6 +158,7 @@ class DatiTrasmissione(models.Model):
                     code="00427")
 
 
+@export
 class Anagrafica(FullNameMixin, models.Model):
     denominazione = fields.StringField(max_length=80, null=True)
     nome = fields.StringField(max_length=60, null=True)
@@ -142,6 +167,7 @@ class Anagrafica(FullNameMixin, models.Model):
     cod_eori = fields.StringField(xmltag="CodEORI", min_length=13, max_length=17, null=True)
 
 
+@export
 class DatiAnagraficiCedentePrestatore(models.Model):
     __xmltag__ = "DatiAnagrafici"
     id_fiscale_iva = IdFiscaleIVA
@@ -152,11 +178,10 @@ class DatiAnagraficiCedentePrestatore(models.Model):
     numero_iscrizione_albo = fields.StringField(max_length=60, null=True)
     data_iscrizione_albo = fields.DateField(null=True)
     regime_fiscale = fields.StringField(
-            length=4, choices=("RF01", "RF02", "RF04", "RF05", "RF06", "RF07",
-                               "RF08", "RF09", "RF10", "RF11", "RF12", "RF13",
-                               "RF14", "RF15", "RF16", "RF17", "RF18", "RF19"))
+            length=4, choices=consts.REGIME_FISCALE)
 
 
+@export
 class IndirizzoType(models.Model):
     indirizzo = fields.StringField(max_length=60)
     numero_civico = fields.StringField(max_length=8, null=True)
@@ -166,10 +191,12 @@ class IndirizzoType(models.Model):
     nazione = fields.StringField(length=2)
 
 
+@export
 class Sede(IndirizzoType):
     pass
 
 
+@export
 class IscrizioneREA(models.Model):
     ufficio = fields.StringField(length=2)
     numero_rea = fields.StringField(xmltag="NumeroREA", max_length=20)
@@ -178,16 +205,19 @@ class IscrizioneREA(models.Model):
     stato_liquidazione = fields.StringField(length=2, choices=("LS", "LN"))
 
 
+@export
 class Contatti(models.Model):
     telefono = fields.StringField(min_length=5, max_length=12, null=True)
     fax = fields.StringField(min_length=5, max_length=12, null=True)
     email = fields.StringField(min_length=7, max_length=256, null=True)
 
 
+@export
 class StabileOrganizzazione(IndirizzoType):
     pass
 
 
+@export
 class CedentePrestatore(models.Model):
     dati_anagrafici = DatiAnagraficiCedentePrestatore
     sede = Sede
@@ -197,6 +227,7 @@ class CedentePrestatore(models.Model):
     riferimento_amministrazione = fields.StringField(max_length=20, null=True)
 
 
+@export
 class DatiAnagraficiCessionarioCommittente(models.Model):
     __xmltag__ = "DatiAnagrafici"
     id_fiscale_iva = fields.ModelField(IdFiscaleIVA, null=True)
@@ -212,6 +243,7 @@ class DatiAnagraficiCessionarioCommittente(models.Model):
                     code="00417")
 
 
+@export
 class RappresentanteFiscale(FullNameMixin, models.Model):
     id_fiscale_iva = fields.ModelField(IdFiscaleIVA, null=True)
     denominazione = fields.StringField(max_length=80, null=True)
@@ -219,6 +251,7 @@ class RappresentanteFiscale(FullNameMixin, models.Model):
     cognome = fields.StringField(max_length=60, null=True)
 
 
+@export
 class CessionarioCommittente(models.Model):
     dati_anagrafici = DatiAnagraficiCessionarioCommittente
     sede = Sede
@@ -226,6 +259,7 @@ class CessionarioCommittente(models.Model):
     rappresentante_fiscale = fields.ModelField(RappresentanteFiscale, null=True)
 
 
+@export
 class DatiAnagraficiRappresentante(models.Model):
     __xmltag__ = "DatiAnagrafici"
     id_fiscale_iva = IdFiscaleIVA
@@ -233,11 +267,13 @@ class DatiAnagraficiRappresentante(models.Model):
     anagrafica = Anagrafica
 
 
+@export
 class RappresentanteFiscaleCedentePrestatore(models.Model):
     __xmltag__ = "RappresentanteFiscale"
     dati_anagrafici = DatiAnagraficiRappresentante
 
 
+@export
 class DatiAnagraficiTerzoIntermediario(models.Model):
     __xmltag__ = "DatiAnagrafici"
     id_fiscale_iva = IdFiscaleIVA
@@ -245,10 +281,12 @@ class DatiAnagraficiTerzoIntermediario(models.Model):
     anagrafica = Anagrafica
 
 
+@export
 class TerzoIntermediarioOSoggettoEmittente(models.Model):
     dati_anagrafici = DatiAnagraficiTerzoIntermediario
 
 
+@export
 class FatturaElettronicaHeader(models.Model):
     dati_trasmissione = DatiTrasmissione
     cedente_prestatore = CedentePrestatore
@@ -258,26 +296,29 @@ class FatturaElettronicaHeader(models.Model):
     soggetto_emittente = fields.StringField(length=2, choices=("CC", "TZ"), null=True)
 
 
+@export
 class DatiRitenuta(models.Model):
-    tipo_ritenuta = fields.StringField(length=4, choices=("RT01", "RT02"))
+    tipo_ritenuta = fields.StringField(length=4, choices=consts.TIPO_RITENUTA)
     importo_ritenuta = fields.DecimalField(max_length=15)
     aliquota_ritenuta = fields.DecimalField(max_length=6)
     causale_pagamento = fields.StringField(max_length=2)
 
 
+@export
 class DatiBollo(models.Model):
     bollo_virtuale = fields.StringField(length=2, choices=("SI",))
     importo_bollo = fields.DecimalField(max_length=15)
 
 
+@export
 class DatiCassaPrevidenziale(models.Model):
-    tipo_cassa = fields.StringField(length=4, choices=["TC{:02d}".format(i) for i in range(1, 23)])
+    tipo_cassa = fields.StringField(length=4, choices=consts.TIPO_CASSA)
     al_cassa = fields.DecimalField(max_length=6)
     importo_contributo_cassa = fields.DecimalField(max_length=15)
     imponibile_cassa = fields.DecimalField(max_length=15)
     aliquota_iva = fields.DecimalField(max_length=6, xmltag="AliquotaIVA")
     ritenuta = fields.StringField(length=2, choices=("SI",), null=True)
-    natura = fields.StringField(length=2, choices=("N1", "N2", "N3", "N4", "N5", "N6", "N7"), null=True)
+    natura = fields.StringField(min_length=2, max_length=4, choices=consts.NATURA_IVA, null=True)
     riferimento_amministrazione = fields.StringField(max_length=20, null=True)
 
     def validate_model(self, validation):
@@ -290,14 +331,16 @@ class DatiCassaPrevidenziale(models.Model):
                 self._meta["natura"], "field has value while aliquota_iva is not zero", code="00414")
 
 
+@export
 class ScontoMaggiorazione(models.Model):
     tipo = fields.StringField(length=2, choices=("SC", "MG"))
     percentuale = fields.DecimalField(max_length=6, null=True)
     importo = fields.DecimalField(max_length=15, null=True)
 
 
+@export
 class DatiGeneraliDocumento(models.Model):
-    tipo_documento = fields.StringField(length=4, choices=("TD01", "TD02", "TD03", "TD04", "TD05", "TD06"))
+    tipo_documento = fields.StringField(length=4, choices=consts.TIPO_DOCUMENTO)
     divisa = fields.StringField()
     data = fields.DateField()
     numero = fields.StringField(max_length=20)
@@ -328,6 +371,7 @@ class DatiGeneraliDocumento(models.Model):
                 self._meta["numero"], "numero must contain at least one number", code="00425")
 
 
+@export
 class AltriDatiGestionali(models.Model):
     tipo_dato = fields.StringField(max_length=10)
     riferimento_testo = fields.StringField(max_length=60, null=True)
@@ -335,11 +379,13 @@ class AltriDatiGestionali(models.Model):
     riferimento_data = fields.DateField(null=True)
 
 
+@export
 class CodiceArticolo(models.Model):
     codice_tipo = fields.StringField(max_length=35)
     codice_valore = fields.StringField(max_length=35)
 
 
+@export
 class DettaglioLinee(models.Model):
     numero_linea = fields.IntegerField(max_length=4)
     tipo_cessione_prestazione = fields.StringField(length=2, choices=("SC", "PR", "AB", "AC"), null=True)
@@ -354,7 +400,7 @@ class DettaglioLinee(models.Model):
     prezzo_totale = fields.DecimalField(max_length=21)
     aliquota_iva = fields.DecimalField(xmltag="AliquotaIVA", max_length=6)
     ritenuta = fields.StringField(length=2, choices=("SI",), null=True)
-    natura = fields.StringField(length=2, null=True, choices=("N1", "N2", "N3", "N4", "N5", "N6", "N7"))
+    natura = fields.StringField(min_length=2, max_length=4, null=True, choices=consts.NATURA_IVA)
     riferimento_amministrazione = fields.StringField(max_length=20, null=True)
     altri_dati_gestionali = fields.ModelListField(AltriDatiGestionali, null=True)
 
@@ -371,10 +417,22 @@ class DettaglioLinee(models.Model):
             validation.add_error(
                 self._meta["natura"], "natura presente a fronte di aliquota_iva diversa da zero", code="00401")
 
+    def autofill_prezzo_totale(self):
+        """
+        Compute prezzo_totale if it was not set
+        """
+        if self.prezzo_totale is not None:
+            return
+        if self.quantita is None:
+            self.prezzo_totale = self.prezzo_unitario
+        else:
+            self.prezzo_totale = self.prezzo_unitario * self.quantita
 
+
+@export
 class DatiRiepilogo(models.Model):
     aliquota_iva = fields.DecimalField(xmltag="AliquotaIVA", max_length=6)
-    natura = fields.StringField(length=2, null=True, choices=("N1", "N2", "N3", "N4", "N5", "N6", "N7"))
+    natura = fields.StringField(min_length=2, max_length=4, null=True, choices=consts.NATURA_IVA)
     spese_accessorie = fields.DecimalField(max_length=15, null=True)
     arrotondamento = fields.DecimalField(max_length=21, null=True)
     # FIXME: Su questo valore il sistema effettua un controllo per verificare
@@ -396,6 +454,7 @@ class DatiRiepilogo(models.Model):
                 self._meta["natura"], "field has value while aliquota_iva is not zero", code="00430")
 
 
+@export
 class DatiBeniServizi(models.Model):
     dettaglio_linee = fields.ModelListField(DettaglioLinee)
     dati_riepilogo = fields.ModelListField(DatiRiepilogo)
@@ -412,16 +471,9 @@ class DatiBeniServizi(models.Model):
         submit a pull request.
         """
         kw.setdefault("numero_linea", len(self.dettaglio_linee) + 1)
-        self.dettaglio_linee.append(DettaglioLinee(**kw))
-
-        # Compute prezzo_totale where not set
-        for d in self.dettaglio_linee:
-            if d.prezzo_totale is not None:
-                continue
-            if d.quantita is None:
-                d.prezzo_totale = d.prezzo_unitario
-            else:
-                d.prezzo_totale = d.prezzo_unitario * d.quantita
+        d = DettaglioLinee(**kw)
+        d.autofill_prezzo_totale()
+        self.dettaglio_linee.append(d)
 
     def build_dati_riepilogo(self):
         """
@@ -443,7 +495,7 @@ class DatiBeniServizi(models.Model):
 
         self.dati_riepilogo = []
         for aliquota, linee in sorted(by_aliquota.items()):
-            imponibile = sum(l.prezzo_totale for l in linee)
+            imponibile = sum(linea.prezzo_totale for linea in linee)
             imposta = imponibile * aliquota / 100
             self.dati_riepilogo.append(
                     DatiRiepilogo(
@@ -451,6 +503,7 @@ class DatiBeniServizi(models.Model):
                         imposta=imposta, esigibilita_iva="I"))
 
 
+@export
 class DatiDocumentiCorrelati(models.Model):
     riferimento_numero_linea = fields.ListField(fields.IntegerField(max_length=4), null=True)
     id_documento = fields.StringField(max_length=20)
@@ -461,26 +514,32 @@ class DatiDocumentiCorrelati(models.Model):
     codice_cig = fields.StringField(max_length=15, xmltag="CodiceCIG", null=True)
 
 
+@export
 class DatiOrdineAcquisto(DatiDocumentiCorrelati):
     pass
 
 
+@export
 class DatiContratto(DatiDocumentiCorrelati):
     pass
 
 
+@export
 class DatiConvenzione(DatiDocumentiCorrelati):
     pass
 
 
+@export
 class DatiRicezione(DatiDocumentiCorrelati):
     pass
 
 
+@export
 class DatiFattureCollegate(DatiDocumentiCorrelati):
     pass
 
 
+@export
 class DatiAnagraficiVettore(models.Model):
     id_fiscale_iva = IdFiscaleIVA
     codice_fiscale = fields.StringField(min_length=11, max_length=16, null=True)
@@ -488,10 +547,12 @@ class DatiAnagraficiVettore(models.Model):
     numero_licenza_guida = fields.StringField(max_length=20, null=True)
 
 
+@export
 class IndirizzoResa(IndirizzoType):
     pass
 
 
+@export
 class DatiTrasporto(models.Model):
     dati_anagrafici_vettore = fields.ModelField(DatiAnagraficiVettore, null=True)
     mezzo_trasporto = fields.StringField(max_length=80, null=True)
@@ -508,6 +569,7 @@ class DatiTrasporto(models.Model):
     data_ora_consegna = fields.DateTimeField(null=True)
 
 
+@export
 class DatiDDT(models.Model):
     __xmltag__ = "DatiDDT"
     numero_ddt = fields.StringField(max_length=20, xmltag="NumeroDDT")
@@ -515,11 +577,13 @@ class DatiDDT(models.Model):
     riferimento_numero_linea = fields.ListField(fields.IntegerField(max_length=4), null=True)
 
 
+@export
 class FatturaPrincipale(models.Model):
     numero_fattura_principale = fields.StringField(max_length=20)
     data_fattura_principale = fields.DateField()
 
 
+@export
 class DatiGenerali(models.Model):
     dati_generali_documento = DatiGeneraliDocumento
     dati_ordine_acquisto = fields.ModelListField(DatiOrdineAcquisto, null=True)
@@ -543,9 +607,10 @@ class DatiGenerali(models.Model):
                 code="00418")
 
 
+@export
 class DettaglioPagamento(models.Model):
     beneficiario = fields.StringField(max_length=200, null=True)
-    modalita_pagamento = fields.StringField(length=4, choices=["MP{:02d}".format(i) for i in range(1, 23)])
+    modalita_pagamento = fields.StringField(length=4, choices=consts.MODALITA_PAGAMENTO)
     data_riferimento_termini_pagamento = fields.DateField(null=True)
     giorni_termini_pagamento = fields.IntegerField(max_length=3, null=True)
     data_scadenza_pagamento = fields.DateField(null=True)
@@ -567,11 +632,13 @@ class DettaglioPagamento(models.Model):
     codice_pagamento = fields.StringField(max_length=60, null=True)
 
 
+@export
 class DatiPagamento(models.Model):
     condizioni_pagamento = fields.StringField(length=4, choices=("TP01", "TP02", "TP03"))
     dettaglio_pagamento = fields.ModelListField(DettaglioPagamento)
 
 
+@export
 class Allegati(models.Model):
     nome_attachment = fields.StringField(max_length=60)
     algoritmo_compressione = fields.StringField(max_length=10, null=True)
@@ -580,11 +647,13 @@ class Allegati(models.Model):
     attachment = fields.Base64BinaryField()
 
 
+@export
 class DatiVeicoli(models.Model):
     data = fields.DateField()
     totale_percorso = fields.StringField(max_length=15)
 
 
+@export
 class FatturaElettronicaBody(models.Model):
     dati_generali = DatiGenerali
     dati_beni_servizi = DatiBeniServizi
@@ -638,6 +707,7 @@ class FatturaElettronicaBody(models.Model):
                     code="00419")
 
 
+@export
 class Fattura(models.Model):
     __xmlns__ = NS
     __xmltag__ = "FatturaElettronica"
@@ -696,6 +766,7 @@ class Fattura(models.Model):
         return super().from_etree(el)
 
 
+@export
 class FatturaPrivati12(Fattura):
     """
     Fattura privati 1.2
@@ -704,6 +775,7 @@ class FatturaPrivati12(Fattura):
         return "FPR12"
 
 
+@export
 class FatturaPA12(Fattura):
     """
     Fattura PA 1.2
@@ -712,7 +784,11 @@ class FatturaPA12(Fattura):
         return "FPA12"
 
 
-def auto_from_etree(root):
+@export
+def auto_from_etree(root: ET.ElementTree) -> Union[Fattura, FatturaElettronicaSemplificata]:
+    """
+    Instantiate a Fattura or FatturaElettronicaSemplificata from a parsed XML
+    """
     from .fattura_semplificata import NS10, FatturaElettronicaSemplificata
     tagname_ordinaria = "{{{}}}FatturaElettronica".format(NS)
     tagname_semplificata = "{{{}}}FatturaElettronicaSemplificata".format(NS10)
@@ -741,3 +817,25 @@ def auto_from_etree(root):
 
     res.from_etree(root)
     return res
+
+
+@export
+def auto_from_dict(data: Dict[str, Any]) -> Union[Fattura, FatturaElettronicaSemplificata]:
+    """
+    Given the equivalent of a Fattura.to_jsonable() data structure,
+    reconstruct the fattura
+    """
+    from .fattura_semplificata import FatturaElettronicaSemplificata
+    try:
+        formato = data["fattura_elettronica_header"]["dati_trasmissione"]["formato_trasmissione"]
+    except KeyError:
+        raise RuntimeError("fattura_elettronica_header.dati_trasmissione.formato_trasmissione not found in input")
+
+    if formato == "FPR12":
+        return FatturaPrivati12(**data)
+    elif formato == "FPA12":
+        return FatturaPA12(**data)
+    elif formato == "FSM10":
+        return FatturaElettronicaSemplificata(**data)
+    else:
+        raise RuntimeError(f"Unsupported formato_trasmissione: {formato!r}")
